# Copyright 2012 Hewlett-Packard Development Company, L.P. All Rights Reserved.
#
# Author: Simon McCartney <simon.mccartney@hp.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import logging
from gitshelf.cli import BaseCommand

LOG = logging.getLogger(__name__)


class GitShelfDiffCommand(BaseCommand):
    """ Check a set of repos for changes"""

    def execute(self, parsed_args):
        """execute, something to do for this command."""
        # load the configuration from yaml, rendering
        # any tokens along the way
        config = self._parse_configuration(parsed_args)

        # get back the collection of books
        books = self._get_books(parsed_args, config)

        # now iterate over the list of book objects
        for book in books:
            book.diff()



# Copyright 2012 Hewlett-Packard Development Company, L.P. All Rights Reserved.
#
# Author: Simon McCartney <simon.mccartney@hp.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from cliff.app import App
from cliff.commandmanager import CommandManager
from gitshelf.version import version_info as version
import logging
import sys


class GitShelfShell(App):
    """Parent class for the 2 sub-commands."""
    log = logging.getLogger(__name__)

    def __init__(self):
        super(GitShelfShell, self).__init__(
            description='Manage a collection of git repos without using submodules',
            version=version.canonical_version_string(),
            command_manager=CommandManager('gitshelf.cli'),
        )

        self.log = logging.getLogger(__name__)

    def build_option_parser(self, description, version):
        parser = super(GitShelfShell, self).build_option_parser(description, version)

        return parser


def main():
    app = GitShelfShell()
    sys.exit(app.run(sys.argv[1:]))



# Copyright 2012 Hewlett-Packard Development Company, L.P. All Rights Reserved.
#
# Author: Simon McCartney <simon.mccartney@hp.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import logging
from gitshelf.cli import BaseCommand
from gitshelf.book import Book

LOG = logging.getLogger(__name__)


class GitShelfDiscoverCommand(BaseCommand):
    """ Discover the git repos & symlinks under this directory """

    def get_parser(self, prog_name):
        parser = super(BaseCommand, self).get_parser(prog_name)
        parser.add_argument('--use-branch',
                            default=False,
                            help="Use the branch name instead of the sha1 for pinning",
                            action='store_true')
        return parser

    def execute(self, parsed_args):
        """execute, something to do for this command."""

        # get back the collection of books
        books = Book.discover(usebranch=parsed_args.use_branch)

        # sort the list, based on the path attribute
        books.sort(key=lambda book: book.path)

        # now iterate over the list of book objects
        # TODO: be less hacky and use a YAML emitter
        print "books:"
        for book in books:
            print "  - book: {0}".format(book.path)
            if book.git is not None:
                print "    git: {0}".format(book.git)
                print "    branch: {0}".format(book.branch)
            elif book.link is not None:
                print "    link: {0}".format(book.link)
            print



# Copyright 2012 Hewlett-Packard Development Company, L.P. All Rights Reserved.
#
# Author: Simon McCartney <simon.mccartney@hp.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from urlparse import urlparse, parse_qsl
from urllib import unquote_plus

def get_item_properties(item, fields, mixed_case_fields=[], formatters={}):
    """Return a tuple containing the item properties.

    :param item: a single item resource (e.g. Server, Tenant, etc)
    :param fields: tuple of strings with the desired field names
    :param mixed_case_fields: tuple of field names to preserve case
    :param formatters: dictionary mapping field names to callables
        to format the values
    """
    row = []

    for field in fields:
        if field in formatters:
            row.append(formatters[field](item))
        else:
            if field in mixed_case_fields:
                field_name = field.replace(' ', '_')
            else:
                field_name = field.lower().replace(' ', '_')
            if not hasattr(item, field_name) and \
                    (isinstance(item, dict) and field_name in item):
                data = item[field_name]
            else:
                data = getattr(item, field_name, '')
            if data is None:
                data = ''
            row.append(data)
    return tuple(row)


def get_columns(data):
    """
    Some row's might have variable count of columns, ensure that we have the
    same.

    :param data: Results in [{}, {]}]
    """
    columns = set()

    def _seen(col):
        columns.add(str(col))

    map(lambda item: map(_seen, item.keys()), data)
    return list(columns)



class Url(object):
    '''A url object that can be compared with other url orbjects
    without regard to the vagaries of encoding, escaping, and ordering
    of parameters in query strings.'''

    def __init__(self, url):
        parts = urlparse(url)
        _query = frozenset(parse_qsl(parts.query))
        _path = unquote_plus(parts.path)
        _path = _path.replace('//', '/')
        parts = parts._replace(query=_query, path=_path)
        self.parts = parts

    def __eq__(self, other):
        return self.parts == other.parts

    def __hash__(self):
        return hash(self.parts)

class NestedDict(dict):
    '''Make accessing nested dictionaries less painful, return
    and empty dict for anything that doesn't exist.
    Stolen from http://ohuiginn.net/mt/2010/07/nested_dictionaries_in_python.html'''

    def __getitem__(self, key):
        if key in self: return self.get(key)
        return self.setdefault(key, NestedDict())



# Copyright 2012 Hewlett-Packard Development Company, L.P. All Rights Reserved.
#
# Author: Simon McCartney <simon.mccartney@hp.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import logging
import yaml
import re
from gitshelf.utils import NestedDict
from gitshelf.book import Book
from cliff.command import Command

LOG = logging.getLogger(__name__)


class BaseCommand(Command):
    """ Parent Command object for gitshelf """

    def get_parser(self, prog_name):
        parser = super(BaseCommand, self).get_parser(prog_name)

        parser.add_argument('--gitshelf',
                            dest='gitshelf',
                            default='gitshelf.yml',
                            nargs=1,
                            help="path to gitshelf YAML config, defaults to gitshelf.yml",
                            action='store')

        parser.add_argument('--token',
                            dest='tokens',
                            default=None,
                            nargs=1,
                            help="inject a token for substituion, overrides any environment level tokens",
                            action='append')

        parser.add_argument('--skip-deletes', help="skip and group/rule deletes", action="store_true")

        parser.add_argument('--fakeroot',
                            dest='fakeroot',
                            default=None,
                            help='path to prepand to any book paths, to test shelves '
                                 'that are absolute without using the absolute path')

        parser.add_argument('--environment',
                            dest='environment',
                            default=None,
                            nargs=1,
                            help='set the desired environment, overriding any default'
                                 'settings in the config file (gitshelf.yml)')

        parser.add_argument('--dry-run',
                            default=False,
                            help="Defaults to False",
                            action='store_true')

        parser.add_argument('--skip-repo-url-check',
                            dest='skiprepourlcheck',
                            default=False,
                            help='Defaults to False'
                                 'Skip the repo remote host check',
                            action='store_true')

        return parser

    def post_execute(self, data):
        """ Format the results locally if needed.

        By default we just return data.

        :param data: Whatever is returned by self.execute()

        """
        return data

    def take_action(self, parsed_args):
        # TODO: Common Exception Handling Here
        results = self.execute(parsed_args)
        return self.post_execute(results)

    def _parse_configuration(self, parsed_args):
        # Read the main config file
        LOG.debug(parsed_args)
        config_file = parsed_args.gitshelf.pop() if isinstance(parsed_args.gitshelf, list) else parsed_args.gitshelf
        LOG.debug("config_file = {0}".format(config_file))

        with open(config_file) as fh:
            config = NestedDict(yaml.load(fh))

        with open(config_file) as fh:
            config_raw = fh.read()

        if parsed_args.environment:
            environment = parsed_args.environment[0]
        else:
            environment = config['defaults'].get('environment', 'dev')

        tokens = config['environments'][environment]['tokens']
        LOG.debug('Tokens: {0}'.format(tokens))

        # overwrite the tokens loaded from the file with any passed on the
        # command line
        if parsed_args.tokens:
            LOG.debug("Tokens have been passed on the command line: {0}".format(parsed_args.tokens))
            # split them into key/value pair, but we are passed a list of lists
            # so cli_token[0] accessess the item we want
            for cli_token in parsed_args.tokens:
                LOG.debug("Parsing the command line token: {0}".format(cli_token))
                split_token = cli_token[0].partition("=")
                LOG.debug("Split command line token: {0}".format(split_token))
                tokens[split_token[0]] = split_token[2]

            LOG.debug('Tokens: {0}'.format(tokens))

        # expand out tokens, tokens are {} wrapped names from the environment section
        delimiters = ('{', '}')

        def _replaceToken(match):
            # Strip the delimiter with 1:-1
            key = match.group(0)[1:-1]
            if key in tokens:
                return tokens[key]    # Strip the delimiter with 1:-1
            return ''

        rendered_config = re.sub('\%s.*?\%s' % delimiters, _replaceToken, config_raw)

        # reload the config block from the rendered yaml
        config = yaml.load(rendered_config)

        LOG.debug(config)

        return config

    def _get_books(self, parsed_args, config):

        LOG.debug("parsed_args: {0}".format(parsed_args))
        LOG.debug("config: {0}".format(config))

        # load the config into an array of Book objects
        books = []
        for book in config['books']:
            LOG.debug("Fresh book: {0}".format(book))
            book.update({'fakeroot': parsed_args.fakeroot})
            LOG.debug("Final book: {0}".format(book))
            # the dictionary we get from the parsed configuration should
            # match the named parameters to the Book class, so we use
            # ** to unpack the dictionary to the class arguments
            books.append(Book(**book))

        return books



# Copyright 2012 Hewlett-Packard Development Company, L.P. All Rights Reserved.
#
# Author: Simon McCartney <simon.mccartney@hp.com>
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import logging
import os
import errno
from sh import git
from gitshelf.utils import Url

LOG = logging.getLogger(__name__)


class Book:
    """ Object to represent a book - repo on disk

        Books can be one of 2 types:
            1) a git repo
            2) a symlink to another location

        Keyword arguments:
            path -- absolute or relative path to the location the repo should be in
            git -- git url to be used as the source for the repo
            branch -- branch/sha1/tag to be used in the checkout, defaults to master
            link -- destination of the link target
            skiprepourlcheck -- flag to skip that the git url matches the definition during status checks
            fakeroot -- a gitshelf may specify absolute paths, setting fakeroot allows you to make an absolute path relative to the passed path

    """

    def __init__(self,
                 book,
                 git=None,
                 branch='master',
                 link=None,
                 skiprepourlcheck=False,
                 fakeroot=None):
        """Instantiate a book object"""
        self.path = book
        self.git = git
        self.link = link
        self.branch = branch
        self.skiprepourlcheck = skiprepourlcheck
        self.fakeroot = fakeroot

        if (self.git is None) and (self.link is None):
            raise StandardError("Book is neither git or link!")

        # Only apply fakeroot to non-relative paths
        if self.fakeroot is not None and self.path[0] != '.':
            LOG.debug('fakepath set, prepending {0} to {1}'.format(
                self.fakeroot, self.path))
            # need to strip any leading os.sep
            book_path = os.path.join(self.fakeroot, self.path.lstrip(os.sep))
            self.path = book_path
            LOG.debug('book.path is now {0}'.format(self.path))

        if self.fakeroot is not None and self.link and self.link[0] == os.sep:
            # Update the link target to, which also means that
            # link targets must now be relative to handle relocating a
            # tarball of the gitshelf.  This is all getting a little
            # complicated.
            link_path = os.path.join(self.fakeroot, self.link.lstrip(os.sep))
            self.link = link_path
            LOG.debug('After fakeroot, book.link is now {0}'.format(self.link))

            # map to a relative path
            link_relative = os.path.relpath(self.link, self.path)
            LOG.debug('book.link ({0}), relative path: {1}'.format(self.link, link_relative))
            self.link = link_relative

    def create(self):
        if self.git and self.link is None:
            self._create_git()
        elif self.link and self.git is None:
            self._create_link()

    def _create_git(self):
        """create a book from a git repo"""

        if not os.path.exists(self.path):
            LOG.info(("Creating book {0} from {1}, branch: {2}" +
                     "").format(self.path, self.git, self.branch))
            git.clone(self.git, self.path)
        else:
            LOG.info("Book {0} already exists".format(self.path))

        cwd = os.getcwd()
        os.chdir(self.path)

        if self.skiprepourlcheck:
            remote_match_found = False
            for remote in git("remote", "-v"):
                remote_parts = remote.split()

                if Url(remote_parts[1]) == Url(self.git):
                    remote_match_found = True

            if remote_match_found:
                LOG.debug('Found {0} in the list of remotes for {1}'.format(self.git, self.path))
            else:
                LOG.error('ERROR: {0} wasn\'t found in the list of remotes for {1}'.format(self.git, self.path))

        if not self._check_branch():
            LOG.info("Switching {0} to branch {1}".format(self.path,
                                                          self.branch))
            git.fetch
            git.checkout(self.branch)

        os.chdir(cwd)

    def _create_link(self):
        """create a book from a link to somewhere else"""

        cwd = os.getcwd()
        parent = os.path.dirname(self.path.rstrip(os.sep))

        if not os.path.islink(self.path):
            LOG.info("Creating book {0} via a link to {1}".format(self.path, self.link))
            # create the parent directory, if required
            self._mkdir_p(os.path.dirname(self.path.rstrip(os.sep)))
            # move to the parent so that creating relative sym-links make sense
            os.chdir(parent)
            # create the symlink
            os.symlink(self.link, os.path.basename(self.path))
        else:
            if not self._check_link():
                LOG.info("Correcting book {0} to {1}".format(self.path, self.link))
                os.remove(self.path)
                # move to the parent so that creating relative sym-links make sense
                os.chdir(parent)
                # re-create the symlink
                os.symlink(self.link, os.path.basename(self.path))
            else:
                LOG.info("Book {0} already exists, target: {1}".format(self.path, os.readlink(self.path)))
        os.chdir(cwd)

    def _mkdir_p(self, path):
        if path == "":
            return

        try:
            os.makedirs(path)
        except OSError as exc:  # Python >2.5
            if exc.errno == errno.EEXIST and os.path.isdir(path):
                pass
            else:
                raise

    def _check_branch(self):
        """Check that the current working directory is at the given branch/sha1"""

        cb = git('describe', '--all', '--contains', '--abbrev=4', 'HEAD').rstrip('\r\n')
        sha1 = git('rev-parse', 'HEAD').rstrip('\r\n')
        LOG.debug("Book {0} should be at {1}".format(self.path, self.branch))
        LOG.debug("Book {0}'s current branch is {1}".format(self.path, cb))
        LOG.debug("Book {0}'s current sha1 is {1}".format(self.path, sha1))

        if ((cb == self.branch) or (sha1 == self.branch)):
            return True
        else:
            LOG.warn("WARNING {0} is at branch:{1} (sha1: {2}), not {3}".format(self.path,
                                                                                cb,
                                                                                sha1,
                                                                                self.branch))
            return False

    def _check_link(self):
        # check the link points to the correct location
        link_target = os.readlink(self.path)
        LOG.debug('book: {0} should point to {1}, it points to {2}'.format(self.path, self.link, link_target))

        if link_target == self.link:
            return True
        else:
            return False

    def status(self):
        if self.git and self.link is None:
            # git repo, check it exists & isn't dirty
            if not os.path.exists(self.path):
                LOG.info("ERROR book {0} from {1} doesn't exist.".format(
                    self.path,
                    self.git))
            else:
                # chdir to the book & run `git status`
                cwd = os.getcwd()
                os.chdir(self.path)
                if self._check_branch():
                    git_status = git.status()
                    if "nothing to commit, working directory clean" in git_status:
                        LOG.info("# book {0} OK".format(self.path))
                    else:
                        LOG.info("# book {0}".format(self.path))
                        LOG.info(git_status)
                os.chdir(cwd)

        elif self.link and self.git is None:
            # check the link points to the correct location
            if self._check_link():
                LOG.info('# book {0} correctly points to {1}'.format(self.path, self.link))
            else:
                link_target = os.readlink(self.path)
                LOG.error('ERROR: {0} should point to {1}, it points to {2}'.format(self.path, self.link, link_target))

        else:
            LOG.error('Unknown book type: {0}'.format(self.path))

    def diff(self):
        if self.git and self.link is None:
            # git repo, check it exists & isn't dirty
            if not os.path.exists(self.path):
                LOG.info("ERROR book {0} from {1} doesn't exist.".format(
                    self.path,
                    self.git))
            else:
                # chdir to the book & run `git status`
                cwd = os.getcwd()
                os.chdir(self.path)
                LOG.info("# book {0}".format(self.path))
                git_diff = git.diff()
                if git_diff:
                    LOG.info("# book {0} had changes:".format(self.path))
                    LOG.info(git_diff)
                else:
                    LOG.info("# book {0} is clean".format(self.path))
                os.chdir(cwd)
        elif self.link and self.git is None:
            # check the link points to the correct location
            link_target = os.readlink(self.path)
            LOG.debug('book: {0} should point to {1}, it points to {2}'.format(self.path, self.link, link_target))
            if link_target == self.link:
                LOG.info('# book {0} correctly points to {1}'.format(self.path, self.link))
            else:
                LOG.error('{0} should point to {1}, it points to {2}'.format(self.path, self.link, self.link))
        else:
            LOG.error('Unknown book type: {0}'.format(self.path))

    def pull(self):
        if self.git and self.link is None:
            # git repo, check it exists & isn't dirty
            if not os.path.exists(self.path):
                LOG.info("ERROR book {0} from {1} doesn't exist.".format(
                    self.path,
                    self.git))
            else:
                # chdir to the book & run `git status`
                cwd = os.getcwd()
                os.chdir(self.path)
                LOG.info("# book {0}".format(self.path))
                git_diff = git.diff()
                if git_diff:
                    LOG.info("# book {0} had changes:".format(self.path))
                    LOG.info(git_diff)
                else:
                    LOG.info("# book {0} is clean".format(self.path))
                os.chdir(cwd)
        elif self.link and self.git is None:
            # check the link points to the correct location
            link_target = os.readlink(self.path)
            LOG.debug('book: {0} should point to {1}, it points to {2}'.format(self.path, self.link, link_target))
            if link_target == self.link:
                LOG.info('# book {0} correctly points to {1}'.format(self.path, self.link))
            else:
                LOG.error('{0} should point to {1}, it points to {2}'.format(self.path, self.link, self.link))
        else:
            LOG.error('Unknown book type: {0}'.format(self.path))

    @staticmethod
    def discover(rootdir='.', usebranch=False):
        """discover all the git repo's under this directory"""
        books = []
        for root, subFolders, files in os.walk(rootdir):
            for file in files:
                this_file = os.path.join(root, file).replace(rootdir + os.sep, '')
                if this_file.endswith('.git/config') and not this_file.startswith('.git/config'):
                    repo = os.path.dirname(os.path.dirname(this_file))
                    branch = (Book._discover_branch(repo))
                    sha1 = (Book._discover_sha1(repo))
                    remotes = Book._discover_remotes(repo)
                    LOG.debug("Found a git repo! {0}".format(repo))
                    LOG.debug("remotes are {0}".format(remotes))
                    LOG.debug("branch is {0}".format(branch))
                    LOG.debug("sha1 is {0}".format(sha1))
                    if usebranch:
                        books.append(Book(book=repo, git=remotes['origin'], branch=branch))
                    else:
                        books.append(Book(book=repo, git=remotes['origin'], branch=sha1))

                if os.path.islink(this_file):
                        books.append(Book(book=this_file, link=os.readlink(this_file)))
        return books

    @staticmethod
    def _discover_branch(path='.'):
        """discover the git branch/sha1 of the given directory"""
        cwd = os.getcwd()
        os.chdir(path)
        cb = git('describe', '--all', '--contains', '--abbrev=4', 'HEAD').rstrip('\r\n')
        os.chdir(cwd)
        return cb

    @staticmethod
    def _discover_sha1(path='.'):
        """discover the git branch/sha1 of the given directory"""
        cwd = os.getcwd()
        os.chdir(path)
        sha1 = git('rev-parse', 'HEAD').rstrip('\r\n')
        os.chdir(cwd)
        return sha1

    @staticmethod
    def _discover_remotes(path='.'):
        """discover the remote repos configured for a repo"""
        cwd = os.getcwd()
        os.chdir(path)
        remotes = {}
        for remote_line in git("remote", "-v"):
            r = remote_line.split()[:2]
            remotes[r[0]] = r[1]
        os.chdir(cwd)
        return remotes

    @staticmethod
    def _discover_remote(path='.'):
        """return  the origin remote, or the first remote if origin isn't defined"""
        remotes = Book._discover_remotes(path)
        if 'origin' in remotes:
            return remotes['origin']
        else:
            return remotes[remotes.keys()[0]]
